open Base
open Exceptions

let with_db ?(read_only = false) filename fn =
  try
    let db =
      match read_only with
      | true -> Sqlite3.db_open ~mode:`READONLY filename
      | false -> Sqlite3.db_open filename
    in
    Exn.protect
      ~f:(fun () ->
        try fn db
        with Sqlite3.Error reason ->
          Sqlite_error (Sqlite3.Rc.ERROR, reason) |> raise)
      ~finally:(fun () -> ignore (Sqlite3.db_close db))
  with Sqlite3.Error reason ->
    let msg = Printf.sprintf "opening %s: %s" filename reason in
    Sqlite_error (Sqlite3.Rc.ERROR, msg) |> raise

let with_stmt db query fn =
  try
    let stmt = Sqlite3.prepare db query in
    Exn.protect
      ~f:(fun () ->
        try fn stmt
        with Sqlite3.Error reason ->
          Sqlite_error (Sqlite3.Rc.ERROR, reason) |> raise)
        (* TODO: would it be better to raise an exception here instead of ignore? *)
      ~finally:(fun () -> ignore (Sqlite3.finalize stmt))
  with Sqlite3.Error reason ->
    let msg = Printf.sprintf "preparing '%s': %s" query reason in
    Sqlite_error (Sqlite3.Rc.ERROR, msg) |> raise

let raise_error db =
  let code = Sqlite3.errcode db in
  let msg = Sqlite3.errmsg db in
  raise (Sqlite_error (code, msg))

let bind_all db stmt params =
  List.iter params ~f:(function name, value ->
      (let idx = Sqlite3.bind_parameter_index stmt name in
       match[@warning "-4"] Sqlite3.bind stmt idx value with
       | Sqlite3.Rc.OK -> ()
       | _ -> raise_error db))

let each_row db stmt fn =
  let rec advance rows =
    match[@warning "-4"] Sqlite3.step stmt with
    | Sqlite3.Rc.ROW ->
        let data = Sqlite3.row_data stmt in
        let headers = Sqlite3.row_names stmt in
        let row = Array.zip_exn headers data in
        advance (Array.append rows [| fn row |])
    | Sqlite3.Rc.DONE -> rows
    | _ -> raise_error db
  in
  advance [||]

let exec_fetch (type t) (module M : From_sql.Sig with type t = t) ?(params = [])
    db query =
  with_stmt db query (fun stmt ->
      if not (List.is_empty params) then bind_all db stmt params;
      each_row db stmt M.of_sql_row)

let exec_only ?(params = []) db query =
  let module Ignore = struct
    type t = unit

    let of_sql_row = ignore
  end in
  ignore (exec_fetch (module Ignore) ~params db query)
