open Exceptions
open Functions
open Core

let create_migrations_table db =
  exec_only db
    {|
    CREATE TABLE IF NOT EXISTS _migrations_
    ( id integer PRIMARY KEY
    , code text NOT NULL
    , created_at text NOT NULL);|}

let fetch_migration db id =
  let open Sqlite3.Data in
  let rows =
    exec_fetch
      (module From_sql.Id)
      db "SELECT code FROM _migrations_ WHERE id = :id;"
      ~params:[ (":id", INT (Int64.of_int id)) ]
  in
  match[@warning "-4"] rows with
  | [| [| ("code", TEXT code) |] |] -> Some code
  | [||] -> None
  | _ -> raise (Corrupted_migration_entry id)

let fail_on_code_mismatch id code_in_db current_code =
  match code_in_db with
  | Some code when not (String.equal code current_code) ->
      raise (Codes_differ (id, code, current_code))
  | _ -> ()

let store_migration db id code =
  let open Sqlite3.Data in
  exec_only db
    {|INSERT OR REPLACE
            INTO _migrations_ (id, code, created_at)
            VALUES (:id, :code, :created_at);|}
    ~params:
      [
        (":id", INT (Int64.of_int id));
        (":code", TEXT code);
        (":created_at", TEXT (Time.to_string (Time.now ())));
      ]

let apply db ~migrations =
  create_migrations_table db;
  List.iteri migrations ~f:(fun id current_code ->
      let code_in_db = fetch_migration db id in
      fail_on_code_mismatch id code_in_db current_code;
      if Option.is_none code_in_db then (
        exec_only db current_code;
        store_migration db id current_code ))

let migrations_table_exists db =
  try
    exec_only db "SELECT id, code, created_at from _migrations_;";
    true
  with Sqlite_error _ -> false

let all_migrations_are_applied db migrations =
  try
    List.for_alli migrations ~f:(fun id current_code ->
        let code_in_db = fetch_migration db id in
        fail_on_code_mismatch id code_in_db current_code;
        Option.is_some code_in_db)
  with Sqlite_error _ | Codes_differ _ -> false

let check db ~migrations =
  migrations_table_exists db && all_migrations_are_applied db migrations
