(** @inline *)
include module type of Exceptions

(** Interface and some implementations of conversions from raw Sqlite
    data to OCaml types *)
module From_sql : sig
  (** Signature of module supporting conversion from Sqlite raw values to a
    particular type *)
  module type Sig = sig
    type t

    (** Converts Sqlite row of (header,data) into [t] or throws
      [Conversion_failed] exception if conversion is not possible *)
    val of_sql_row : (string * Sqlite3.Data.t) array -> t
  end

  (** Identity converter. Returns raw Sqlite data unchanged. Never fails. *)
  module Id : Sig with type t = (string * Sqlite3.Data.t) array

  (** Converts single column INT result to [int64] *)
  module Int64 : Sig with type t = int64

  (** Converts single column FLOAT result to [float] *)
  module Float : Sig with type t = float

  (** Converts single column BLOB or TEXT result to [string] *)
  module String : Sig with type t = string
end

(** Database migration helper functions *)
module Migration : sig
  (** Applies all pending migrations to the database in the order
      they've been given.

      Stores the code of all applied migrations in [_migrations_] table.
      Does not apply a migration if it has already been applied and its
      current code matches the one stored in the database.

      Raises an exception whenever SQLite does not return OK.
      Raises an exception when the code of an already applied migration
      under the index [i] differs from the code in [migrations(i)].
      Re-raises all other underlying exceptions from Sqlite3.
  *)
  val apply : Sqlite3.db -> migrations:string list -> unit

  (** Checks that all migrations are applied to the database. *)
  val check : Sqlite3.db -> migrations:string list -> bool
end

(** [with_db ?read_only filename fn] opens a database from [filename], applies
      a callback [fn] to it and closes it after [fn] returns *)
val with_db : ?read_only:bool -> string -> (Sqlite3.db -> 'a) -> 'a

(** [exec_only ?params db query] creates a new Sqlite statement from [query],
      binds [?params] to it and executes it without returning any result *)
val exec_only :
  ?params:(string * Sqlite3.Data.t) list -> Sqlite3.db -> string -> unit

(** [exec ?params db query] creates a new Sqlite statement from [query],
      binds [?params] to it and executes it reading every resulting row and
      converting it using [of_sql_row] function from the given module. *)
val exec_fetch :
  (module From_sql.Sig with type t = 't) ->
  ?params:(string * Sqlite3.Data.t) list ->
  Sqlite3.db ->
  string ->
  't array
